# Descriptive notes for each new metadata version. These are stored in a Python dict
# rather than as a comment, so that the existence of a description can be enforced by
# the unit tests when new versions are added.
VERSION_NOTES = {
    31: 'Add table ids to metadata structs',
    30: 'Store default values and constant arguments as literals',
    29: 'Add user and additional_md fields to metadata structs',
    28: 'Enable view creation from DataFrame with select clause',
    27: 'Enable pxt.query parameterization of limit clauses',
    26: 'Rename clip_text and clip_image to clip',
    25: 'Functions with multiple signatures',
    24: 'Added TableMd/IndexMd.indexed_col_tbl_id',
    23: 'DataFrame.from_clause',
    22: 'TableMd/ColumnMd.media_validation',
    21: 'Separate InlineArray and InlineList',
    20: 'Store DB timestamps in UTC',
    19: 'UDF renames; ImageMemberAccess removal',
    18: 'Restructured index metadata',
    17: 'Renamed remotes to external_stores',
    16: 'Query functions; deferred Expr deserialization',
    15: 'Remotes in table metadata',
}
